# Cours 11 - 1I001 - UPMC - "Jeu en Python"
# Romain Demangeon
# Le fichier graphics.py (qui depend du module tkinter) est necessaire.
# Les primitives (hors-programme) utilisees dans graphics.py sont lastKey (la derniere touche pressee), GraphWin (une fenetre graphique),
#     Point, Rectangle, Image (une sprite), Text (un affichage de texte).
# le jeu se lance avec main()

from graphics import *
import tkinter as old_tk
import random, time

# colonnes : int
colonnes = 20 # nombre de colonnes de l'aire de jeu
# lignes : int
lignes = 14 # nombre de lignes de l'aire de jeu
# nb_murs : int
nb_murs = 25 # nombre de murs
# nb_monstres : int
nb_monstres = 5 # nombre de monstres
# nb_vies : int
nb_vies = 3 # nombre de vies

def fait_tour(etat, fen):
    """ Etat * GraphWin -> bool
    Gere les dessins et les deplacement d'un tour de jeu
    renvoie True si le jeu n'est ni perdu ni gagne"""
    
    if not etat.get_gagne() and not etat.get_perdu():
        # e : Objet
        for e in etat.get_obs(): # pour chaque objet de l'etat
            e.dessine(fen) # dessine l'objet
            if isinstance(e, Mob):
                e.deplace(etat) # le deplace s'il est mobile
        return True
    else:
        return False

def main():
    """ -> Nonetype
    Lance le jeu"""

    # fenetre_du_jeu : GraphWin
    fenetre_du_jeu = GraphWin('Jeu 1I001', colonnes * 32, (lignes + 1) * 32)
    # etat_du_jeu : Etat
    etat_du_jeu = Etat()

    # i : int
    for i in range(nb_murs): # Creation des murs
        m = Mur(random.randint(0, colonnes-1),random.randint(0, lignes-1))
        etat_du_jeu.ajoute_ob(m)

    # monstres : int
    monstres = 0
    while monstres < nb_monstres: # Creation des monstres
        # x, y : int, int
        x, y = (random.randint(0, colonnes-1),random.randint(0, lignes-1)) # coordonnees aléatoires
        if est_vide(etat_du_jeu, x, y): # si les coordonnees sont vides, ajoute un monstre
            m = Monstre(x, y)
            etat_du_jeu.ajoute_ob(m)
            monstres += 1

    # joueurs : int
    joueurs = 0
    while joueurs < 1: # Creation du joueur
        x, y = (random.randint(0, colonnes-1),random.randint(0, lignes-1)) # coordonnees aléatoires
        if est_vide(etat_du_jeu, x, y): # si les coordonnees sont vides, ajoute un joueur
            m = Joueur(x, y, nb_vies)
            etat_du_jeu.ajoute_ob(m)
            m.lier(fenetre_du_jeu)
            joueurs += 1
    
    while fait_tour(etat_du_jeu, fenetre_du_jeu): # boucle principale du jeu
        print("tour", etat_du_jeu.compte())
        time.sleep(0.1) # 100 ms entre chaque 'tour'

    if etat_du_jeu.get_perdu(): # si le jeu finit en étant perdu
        # s : str
        s = "PERDU en " + str(etat_du_jeu.compte()) + " tours."
        # t : Text
        t = Text(Point((colonnes - 4) * 32, lignes * 32 + 16), s) # le texte a afficher
        t.draw(fenetre_du_jeu)
        print("PERDU")

    if etat_du_jeu.get_gagne():
        # s : str
        s = "GAGNE en " + str(etat_du_jeu.compte()) + " tours."
        # t : Text
        t = Text(Point((colonnes - 4) * 32, lignes * 32 + 16), s) # le texte a afficher
        t.draw(fenetre_du_jeu)

def est_vide(etat, x, y):
    """ Etat * int * int -> bool
    Decide si la case (x,y) de l'etat est vide."""

    # e : Objet
    for e in etat.get_obs(): # parcours des objets de l'etat
        if (x, y) == e.get_pos():
            return False
    return True

def est_en(etat, x, y):
    """ Etat * int * int -> bool
    Renvoie l'objet present dans la case (x,y) s'il existe."""

    # e : Objet
    for e in etat.get_obs(): # parcours des objets de l'etat
        if (x, y) == e.get_pos():
            return e
    return None

def est_dans_cadre(etat, x, y):
    """ Etat * int *int
    Decide si la case (x,y) n'est pas hors-limite."""

    return (0 <= x < colonnes) and (0 <= y < lignes)

class Etat:
    """ Represente l'etat du jeu."""
    
    def __init__(self):
        """ self ->  Nonetype """
        # self._obs : list[Objet]
        self._obs = [] # liste des objets de l'etat
        # self._gagne : bool
        self._gagne = False # True si le jeu est gagne
        # self._perdu : bool
        self._perdu = False # True si le jeu est perdu
        # self._tour : int
        self._tour = 0 # tour courant du jeu

    def compte(self):
        """ self -> int
        Donne le tour courant et incrémente ce nombre."""
        
        self._tour += 1
        return self._tour

    def get_obs(self):
        """ self -> list[Object]
        Renvoie la liste des objets."""

        return self._obs

    def ajoute_ob(self, ob):
        """ self * Objet -> Nonetype
        Ajoute un objet a l'etat. """
        
        self._obs.append(ob)

    def enleve_ob(self, ob):
        """ self * Objet -> Nonetype
        Enleve un objet a l'etat. """

        # R : list[Objet]
        R =[] # liste resultat

        # e : Objet
        for e in self._obs:
            if e != ob:
                R.append(e)
        self._obs = R

    def set_gagne(self):
        """ self -> Nonetype
        Met l'etat du jeu en victoire."""
        
        self._gagne = True

    def get_gagne(self):
        """ self -> bool
        Decide si l'etait est une victoire."""
        
        return self._gagne

    def set_perdu(self):
        """ self -> Nonetype
        Met l'etat du jeu en defaite."""
        
        self._perdu = True

    def get_perdu(self):
        """ self -> bool
        Decide si l'etait est une defaite."""
        
        return self._perdu
    
class Ob:
    """ Un objet du jeu. """

    def __init__(self, x, y):
        """ self * int * int -> Nonetype """
        self._coord = (x, y) # coordonnees de l'objet

    def dessine(self, fen):
        """ self * GraphWin -> Nonetype
        Dessine l'objet dans la fenetre fen."""
        
        return None # par defaut, ne dessine rien

    def get_pos(self):
        """ self -> tuple(int, int)
        Retourne la position de l'objet."""
        
        return self._coord 

    def set_pos(self, x, y):
        """ self * int * int -> Nonetype
        Deplace l'objet dans la case (x, y)"""
        
        self._coord = (x, y)

class Mur(Ob):
    """ Un mur inanime. """
    
    def dessine(self, fen):
        """ <Redefinition> """

        x, y = self.get_pos()
        rect = Rectangle(Point(x*32, y*32),Point(x*32 + 31, y*32 + 31))
        rect.setFill('black')
        rect.draw(fen)

class Mob(Ob):
    """ Un objet mobile. """

    def __init__(self, x, y):
        """ <Redefinition> """
        
        Ob.__init__(self, x, y)
        self._dessin = None # l'image qui represente l'objet
   
    def deplace_gauche(self, etat):
        """ self * Etat -> NoneType
        Deplace l'objet vers la gauche si c'est possible."""

        x, y = self.get_pos()
        if est_vide(etat, x-1, y) and est_dans_cadre(etat, x-1, y): # se deplace si la case est vide et existe
            self.set_pos(x-1,y)

    def deplace_droite(self, etat):
        """ self * Etat -> NoneType
        Deplace l'objet vers la droite si c'est possible."""

        x, y = self.get_pos()
        if est_vide(etat, x+1, y) and est_dans_cadre(etat, x+1, y): # se deplace si la case est vide et existe
            self.set_pos(x+1,y)
    
    def deplace_bas(self, etat):
        """ self * Etat -> NoneType
        Deplace l'objet vers le bas si c'est possible."""

        x, y = self.get_pos()
        if est_vide(etat, x, y+1) and est_dans_cadre(etat, x, y+1): # se deplace si la case est vide et existe
            self.set_pos(x,y+1)

    def deplace_haut(self, etat):
        """ self * Etat -> NoneType
        Deplace l'objet vers le haut si c'est possible."""

        x, y = self.get_pos()
        if est_vide(etat, x, y-1) and est_dans_cadre(etat, x, y-1): # se deplace si la case est vide et existe
            self.set_pos(x,y-1)

    def deplace(self, etat):
        """ self * Etat -> Nonetype
        Execute la strategie de deplacement de l'objet mobile."""

        return None # par defaut, pas de strategie de deplacement


class Monstre(Mob):
    """ Un Monstre qui se deplace aléatoirement et peut attaquer un Joueur."""

    def __init__(self, x, y):
        """ <Redefinition> """
        
        Mob.__init__(self, x, y)
        self._vivant = True # Decide si le monstre est vivant.

    def dessine(self, fen):
        """ <Redefinition> """
        
        if self._dessin != None: # si l'objet a deja ete dessine, on l'efface.
            self._dessin.undraw()
        if self._vivant: # si le monstre est vivant, il affiche son sprite a ses coordonnees.
            x, y = self.get_pos()
            self._dessin = Image(Point(x * 32 + 16, y * 32 + 16), "monstre.gif")
            self._dessin.draw(fen)

    def deplace(self, etat):
        """ <Redefinition> """
        
        if self._vivant: # si le monstre est vivant
            x, y = self.get_pos()

            # j : Joueur
            j = None
            
            for dx in {-1, 0, 1}: # parcours de toutes les cases autour du monstres
                for dy in {-1, 0, 1}:
                    if isinstance(est_en(etat, x+dx, y+dy), Joueur): # si un Joueur est trouve, le met dans j
                        j = est_en(etat, x+dx, y+dy)
            if j != None and (random.randint(0,1) == 0): # si un joueur a ete trouve, une chance sur deux de le toucher
                j.touche()
            else:
                # r : int
                r = random.randint(1, 8) # si aucun joueur n'est trouve, une chance sur deux de faire un deplacement aleatoire
                if r == 1:
                    self.deplace_gauche(etat)
                elif r == 2:
                    self.deplace_droite(etat)
                elif r == 3:
                    self.deplace_haut(etat)
                elif r == 4:
                    self.deplace_bas(etat)
        else: # si le monstre est mort, on l'enleve de l'etat
            etat.enleve_ob(self)
            # nb : int
            nb = 0
            for e in etat.get_obs():
                if isinstance(e, Monstre):
                    nb += 1
            if nb == 0:
                etat.set_gagne()

    def touche(self):
        """ self -> Nonetype
        Passe l'etat du monstre a 'mort' (appelee quand le monstre est touche)"""

        self._vivant = False

class Joueur(Mob):
    """ Un Joueur qui se deplace selon les commandes clavier. """
    
    def __init__(self, x, y, n):
        """ <Redefinition> """
        
        Mob.__init__(self, x, y)
        self._invincible = 0 # Nombre de tours ou le joueur sera encore invincible (apres avoir ete touche)
        self._vie = n # Vies restantes au joueur
        self._coeurs = [] # Liste d'images de coeur representant les vies du joueur
        self._fen = None # La fenetre associee au joueur
        
        # i : int
        for i in range(self._vie):
            # im : Image
            im = Image(Point(i * 32 + 16, lignes * 32 + 16), "coeur.gif")
            self._coeurs.append(im)            
        
    def lier(self, fen):
        """ self * GraphWin -> NoneType
        Associe une fenetre au joueur (necessaire pour lier les touches pressees et le deplacement du joueur)."""

        self._fen = fen    
        
    def dessine(self, fen):
        """ <Redefinition> """
        
        if self._dessin != None: # si l'objet a deja ete dessine, on l'efface.
            self._dessin.undraw()
        x, y = self.get_pos()            
        if self._invincible>0: # si le joueur est encore invincible, on affiche un rectangle rouge a la place
            self._dessin = Rectangle(Point(x * 32, y * 32), Point(x * 32 + 31, y * 32 + 31))
            self._dessin.setFill('red')
        else: # sinon on affiche le sprite du joueur
            self._dessin = Image(Point(x * 32 + 16, y * 32 + 16), "joueur.gif")
        self._dessin.draw(fen)
        for e in self._coeurs: # on efface les coeurs
            e.undraw()
        for i in range(self._vie): # on en affiche autant que necessaire
            self._coeurs[i].draw(fen)
            
    def deplace(self, etat):
        """ <Redefinition> """
        
        if self._vie < 1: # si le joueur n'a plus de vie, le jeu est perdu.
            etat.set_perdu()
        if self._invincible > 0: # si le joueur est encore invincible, on décrémente son compteur d'invincibilite.
            self._invincible -= 1
        x, y = self.get_pos()
        if (self._fen).lastKey == 'Left': # si la derniere touche pressee est 'gauche'
            print('deplacement gauche')
            # m : Objet
            m = est_en(etat, x-1, y)
            if isinstance(m, Monstre): # si l'objet a gauche est un Monstre, on le touche
                m.touche()
            else: # sinon on se deplace a gauche
                self.deplace_gauche(etat) 
            (self._fen).lastKey = '' # on oublie la derniere touche pressee
        elif (self._fen).lastKey == 'Right': # si la derniere touche pressee est 'droite'
            print('deplacement droite')
            m = est_en(etat, x+1, y)
            if isinstance(m, Monstre): # si l'objet a droite est un Monstre, on le touche
                m.touche()
            else: # sinon on se deplace a droite
                self.deplace_droite(etat)
            (self._fen).lastKey = '' # on oublie la derniere touche pressee
        elif (self._fen).lastKey == 'Up': # si la derniere touche pressee est 'haut'
            print('deplacement haut')
            m = est_en(etat, x, y-1)
            if isinstance(m, Monstre): # si l'objet en haut est un Monstre, on le touche
                m.touche()
            else: # sinon on se deplace en haut
                self.deplace_haut(etat)
            (self._fen).lastKey = '' # on oublie la derniere touche pressee
        elif (self._fen).lastKey == 'Down': # si la derniere touche pressee est 'bas'
            print('deplacement bas')
            m = est_en(etat, x, y+1)
            if isinstance(m, Monstre): # si l'objet en bas est un Monstre, on le touche
                m.touche()
            else: # sinon on se deplace en bas
                self.deplace_bas(etat)
            (self._fen).lastKey = '' # on oublie la derniere touche pressee

    def touche(self):
        """ self -> Nonetype
        Passe le joueur en invincibilite et lui enleve une vie quand il est touche."""
        
        if self._invincible == 0:
            print("touche")
            self._invincible = 5
            self._vie -= 1
                

# ne lance le jeu que si on interprete directement ce fichier
if __name__ == "__main__":
    main()
